# -*- coding: utf-8 -*-
"""
Utility Functions
=================

**Summary** Utility functions for KSPies.

  :References:

    .. [Mirko2014] Mirko Franchini, Pierre Herman Theodoor Philipsen, Erik van Lenthe, and Lucas Visscher.
        Accurate Coulomb Potentials for Periodic and Molecular Systems through Density Fitting. (2014)
        <https://doi.org/10.1021/ct500172n> Journal of Chemical Theory and Computation, 10(5), 1994-2004.

.. moduleauthor::
    Seungsoo Nam <skaclitz@yonsei.ac.kr> <http://tccl.yonsei.ac.kr/mediawiki/index.php/Main_Page> ORCID: `000-0001-9948-6140 <https://orcid.org/0000-0001-9948-6140>`_
    Hansol Park <hansol7954@yonsei.ac.kr> <http://tccl.yonsei.ac.kr/mediawiki/index.php/Main_Page> ORCID: `000-0001-8706-5015 <https://orcid.org/0000-0001-8706-5015>`_

.. topic:: functions

    mo2ao

    wfnreader

    eval_vh

    eval_vxc

.. todo::

    * Make IF statment for kspies_fort

.. topic:: Internal Log

    **2020-06-06** SN made edits

    **2020-07-26** Updates on eval_vh (Hansol ver.)

    **2020-08-21** SN corrected typos, minor changes in attribute names, etc.

    **2020-11-11** Added wfnreader

"""
from functools import reduce
import numpy as np
import warnings

from scipy.special import sph_harm
from scipy.spatial import distance_matrix
from scipy.interpolate import CubicSpline
from pyscf import gto, dft, scf
from pyscf.dft import numint
try:
    from kspies import kspies_fort
    kf_imported=True
except:
    kf_imported=False

def mo2ao(mo, p1, p2=None):
    """Summary: Convert mo-basis density matrices to basis-set representation density matrices

        Args:
            mo (ndarray) : molecular orbital coefficients
            p1 (ndarray) : mo-basis one-particle density matrices
            p2 (ndarray) : mo-basis two-particle density matrices, optional

        Returns:
            (tuple): tuple containing:

                (ndarray): **dm1** ao-basis one-particle density matrices

                (ndarray): **dm2** ao-basis two-particle density matrices, returned only when p2 is given
    """
    def _convert_rdm1(mo, p1):
        """ Summary: Convert mo-basis 1-rdm p1 to ao-basis
        """
        return reduce(np.dot, (mo, p1, mo.T))
    def _convert_rdm2(mo1, mo2, p2):
        """ Summary: Convert mo-basis 2-rdm p2 to ao-basis
        """
        nmo = mo1.shape[1]
        p = np.dot(mo1, p2.reshape(nmo, -1))
        p = np.dot(p.reshape(-1, nmo), mo2.T)
        p = p.reshape([nmo]*4).transpose(2, 3, 0, 1)
        p = np.dot(mo2, p.reshape(nmo, -1))
        p = np.dot(p.reshape(-1, nmo), mo1.T)
        p = p.reshape([nmo]*4)
        return p

    if np.array(p1).ndim == 2: #RHF
        dm1 = _convert_rdm1(mo, p1)
        if p2 is None:
            return dm1
        dm2 = _convert_rdm2(mo, mo, p2)
        return dm1, dm2
    elif np.array(p1).ndim == 3:
        if np.array(mo).ndim == 2: #ROHF
            mo_a = mo
            mo_b = mo
        else: #UHF
            mo_a, mo_b = mo
        dm1a = _convert_rdm1(mo_a, p1[0])
        dm1b = _convert_rdm1(mo_b, p1[1])
        if p2 is None:
            return (dm1a, dm1b)
        dm2aa = _convert_rdm2(mo_a, mo_a, p2[0])
        dm2ab = _convert_rdm2(mo_a, mo_b, p2[1])
        dm2bb = _convert_rdm2(mo_b, mo_b, p2[2])
        return (dm1a, dm1b), (dm2aa, dm2ab, dm2bb)

TYPE_MAP = [
    [1],  # S
    [2, 3, 4],  # P
    [5, 8, 9, 6, 10, 7],  # D
    [11,14,15,17,20,18,12,16,19,13],  # F
    [21,24,25,30,33,31,26,34,35,28,22,27,32,29,23],  # G
    [56,55,54,53,52,51,50,49,48,47,46,45,44,43,42,41,40,39,38,37,36],  # H
]

def readwfn(filename, mol, makerdm=False):
    """Summary: Convert wfn file to PySCF format

        Args:
            filename (string) : name of wfn file to convert
            mol (object) : an instance of :class:`Mole`

        Returns:
            (tuple): tuple containing:

                (ndarray): **mo_coeff** molecular orbital coefficient from wfn

                (ndarray): **mo_occ** molecular orbital occupation number from wfn

                (ndarray): **mo_energy** molecular orbital energy from wfn

    """

    if mol.cart:
        raise NotImplementedError('Cartesian basis not available')
      
    with open(filename, 'r') as f:
        f.readline()
        dat = f.readline().split()
        norb = int(dat[-7])
        nprm = int(dat[-4])
        natm = int(dat[-2])
        achrg = [] #For sanity check
        coord = []
        for i in range(natm):
            dat = f.readline().split()
            achrg.append(float(dat[-1]))
            coord.append([float(c) for c in dat[4:7]])
        chrg_criteria = sum(abs(mol.atom_charges() - np.array(achrg))) > 1e-10
        coord_criteria = np.linalg.norm(mol.atom_coords() - np.array(coord)) >  1e-6
        if chrg_criteria or coord_criteria:
            warnings.warn("Different molecule!")

        centr = []
        typea = []
        expos = []
        for i in range(nprm//20+1):
            centr += [int(a) for a in f.readline().split()[2:]]
        for i in range(nprm//20+1):
            typea += [int(a) for a in f.readline().split()[2:]]
        for i in range(nprm//5+1):
            expos += [float(a.replace('D', 'E')) for a in f.readline().split()[1:]]

        MOs = []
        occ = []
        eng = []
        for n in range(norb):
            dat = f.readline().split()
            eng.append(float(dat[-1]))
            occ.append(float(dat[-5]))
            orb = []
            for i in range(nprm//5+1):
                orb += [float(a.replace('D','E')) for a in f.readline().split()]
            MOs.append(orb)
    MOs = np.array(MOs).T

    c2s=[]
    for l in range(5):
        c2s.append(np.linalg.pinv(gto.cart2sph(l)))
    from pyscf.x2c import x2c
    uncmol, ctr = x2c._uncontract_mol(mol, True, 0.)
    uncmol_info = []
    for ib in range(uncmol.nbas):
        ia = uncmol.bas_atom(ib)
        l = uncmol.bas_angular(ib)
        es = uncmol.bas_exp(ib)
        uncmol_info.append([ia+1, l, es[0]])

    match = []
    for ip in range(nprm):
        am = [l for l in range(5) if typea[ip] in TYPE_MAP[l]]
        match.append(uncmol_info.index([centr[ip], am, expos[ip]]))

    Rot = np.zeros((uncmol.nao_nr(), nprm))
    bidx = 0
    for ib in range(uncmol.nbas):
        l = uncmol.bas_angular(ib)
        indices = [i for i, ip2b in enumerate(match) if ip2b==ib]
        matchtype = [typea[idx] for idx in indices]
        reorder = [TYPE_MAP[l].index(i) for i in matchtype]
        trans = c2s[l]*1./float(uncmol._libcint_ctr_coeff(ib))
        Rot[bidx:bidx+2*l+1, indices] = trans[:, reorder]
        bidx += 2*l+1

    #Outputs
    cof = np.linalg.pinv(ctr)@Rot@MOs
    eng = np.array(eng)
    occ = np.array(occ)

    restri = False
    isvirt = False
    if max(occ) > 1.3: 
        #Assuming NO occupation numbers of unrestricted calculation 
        #does not exceed 1.3
        restri = True
    if restri and norb > mol.nelectron//2:
        isvirt = True
    elif not restri and norb > mol.nelectron:
        isvirt = True

    s1e = mol.intor_symmetric('int1e_ovlp')
    if restri:
        if isvirt:
            mo_occ = occ
            mo_energy = eng
            mo_coeff = cof
        else:
            mo_occ = np.zeros((mol.nao_nr()))
            mo_energy = np.zeros((mol.nao_nr()))
            mo_coeff = np.zeros((mol.nao_nr(), mol.nao_nr()))
            mo_occ[:len(occ)] = occ
            mo_energy[:len(eng)] = eng
            mo_coeff[:, :len(occ)] = cof
        chk = np.einsum('ki,kl,lj->ij', cof, s1e, cof, optimize='greedy')
        condi = np.linalg.norm(chk - np.eye(len(chk[0])))
        if makerdm:
            dm = scf.hf.make_rdm1(mo_coeff, mo_occ)
    else: #assuming orbital order (alpha_0, alpha_1, ... beta_0, beta_1 ...)
        if isvirt:
            na = norb//2
            mo_occ = np.array([occ[:na], occ[na:]])
            mo_energy = np.array([eng[:na], eng[na:]])
            mo_coeff = np.array([cof[:, :na], cof[:, na:]])
        else:
            na, nb = mol.nelec
            if not na+nb == norb:
                warnings.warn("Proper number of electron should be given from Mole object")
                return cof, occ, eng
            mo_occ = np.zeros((2, mol.nao_nr()))
            mo_energy = np.zeros((2, mol.nao_nr()))
            mo_coeff = np.zeros((2, mol.nao_nr(), mol.nao_nr()))
            mo_occ[0, :na] = occ[:na]
            mo_occ[1, :nb] = occ[na:]
            mo_energy[0, :na] = eng[:na]
            mo_energy[1, :nb] = eng[na:]
            mo_coeff[0,: ,:na] = cof[:, :na]
            mo_coeff[1,: ,:nb] = cof[:, na:]
        chk_a = np.einsum('ki,kl,lj->ij', cof[:, :na], s1e, cof[:, :na], optimize='greedy')
        chk_b = np.einsum('ki,kl,lj->ij', cof[:, na:], s1e, cof[:, na:], optimize='greedy')
        condi = np.linalg.norm(chk_a - np.eye(len(chk_a[0])))
        condi += np.linalg.norm(chk_b - np.eye(len(chk_b[0])))
        if makerdm:
            dm = scf.uhf.make_rdm1(mo_coeff, mo_occ)

    if condi > 1e-5:
        print("Orthonrmal conditonal number:", condi, "> 1e-5")
        warnings.warn("Converted MOs are not orthonormal")
    if makerdm:
        return mo_coeff, mo_occ, mo_energy, dm
    else:
        return mo_coeff, mo_occ, mo_energy


def parse_guide(description):
    """Summary: Guiding potential parser for ZMP and WY

        Args:
            description (str) : guiding potential description for inversion

        Returns:
            (tuple): tuple containing:

                (float): **fac_faxc** factor for Fermi-Amaldi potential (faxc)

                (string): **dft_xc** description of dft part of xc

    """
    def _parse_guide(description):
        fac_faxc = 0
        dftxc = ''
        for token in description.replace('-', '+-').replace(';+', ';').split('+'):
            if token[0] == '-':
                sign = -1
                token = token[1:]
            else:
                sign = 1

            if '*' in token:
                fac, key = token.split('*')
                if fac[0].isalpha():
                    fac, key = key, fac
                fac = sign * float(fac)
            else:
                fac, key = sign, token
            if key.lower() == 'faxc':
                fac_faxc += fac
            else:
                dftxc += '+'+str(fac)+'*'+key
        return fac_faxc, dftxc[1:]

    if ',' in description:
        x_code, c_code = description.split(',')
        fx,dft_x = _parse_guide(x_code)
        fc,dft_c = _parse_guide(c_code)
        fac_faxc = fx + fc
        dft_xc = dft_x + ',' + dft_c
    else:
        fac_faxc, dft_xc = _parse_guide(description)
    return fac_faxc, dft_xc

#controller
radi_method = dft.radi.gauss_chebyshev

def eval_vh(mol, coords, dm, Lvl=3, ang_lv=2): #only atom dependent values
    """Summary: Calculate real-space Hartree potential from given density matrix. See [Mirko2014]_ for some extra context.

        Args:
            mol (object) : an instance of :class:`Mole`
            coords (ndarray) :  grids space used for calculating Hartree potential
            dm (ndarray) : one-particle reduced density matrix in basis set representation
            Lvl (integer) : Interpolation grids space level (input : 0 ~ 9, default 3)
            ang_lv (integer) : setting a limit of angular momentum of spherical harmonics (input : 0 ~ 4, default 2)

        Returns:
            (ndarray): **vh** Pointwise Hartree potential
    """
    def _Cart_Spharm(xyz, lmax):
        """Summary: Cartesian spherical harmonics Z for given xyz coordinate

        Args:
            xyz (ndarray) : 3D coordinates to calculate spherical harmonics
            lmax (integer) : Maximum angular momentum quantum number to calculate spherical harmonics

        Returns:
            (array): from m = -l to l, return array Z of shape (ncoord, lmax+1, 2*lmax+1)
                for specific l & m, call Z[:, l, lmax+m]
        """
        ncoord = np.size(xyz[:, 0])
        rho = np.zeros((ncoord)) #distance from origin
        azi = np.zeros((ncoord)) #azimuth angle, theta in scipy
        pol = np.zeros((ncoord)) #polar angle, phi in scipy
        Z = np.zeros((ncoord, (lmax+1)**2))   #Real spherical harmonics

        xy = xyz[:, 0]**2 + xyz[:, 1]**2
        rho = np.sqrt(xy + xyz[:, 2]**2)
        azi = np.arctan2(xyz[:, 1], xyz[:, 0])
        pol = np.arctan2(np.sqrt(xy), xyz[:, 2])

        a = np.sqrt(0.5)
        for l in range(lmax+1):
            for m in range(1, l+1):
                Yp = sph_harm(m, l, azi, pol)
                Yn = sph_harm(-m, l, azi, pol)
                Z[:, l*(l+1)-m] = a*np.real(1.j*(Yn-((-1.)**m)*Yp))
                Z[:, l*(l+1)+m] = a*np.real((Yn+((-1.)**m)*Yp))

            Z[:, l*(l+1)] = np.real(sph_harm(0, l, azi, pol))

        return Z

    def _grid_refine(n_rad, n_ang, grid):
        """Summary: Reorder grids generated by PySCF for easy handling

        Args:
            n_rad (integer) : the number of radial grid
            n_ang (integer) : the number of angular grid
            grid (ndarray) : 1D sliced grid info (x, y, z coordinates or weights) generated by PySCF

        Returns:
            (ndarray): **Reordered gridpoints** n_ang grids belonging to the same radial grid
        """
        nrest = n_rad%12
        nmain = int((n_rad-nrest)/12)

        m = grid[:nmain*12*n_ang]
        m = m.reshape(nmain, n_ang, 12)
        mmain = np.zeros((nmain*12, n_ang))
        for i in range(nmain):
            mmain[i*12:(i+1)*12, :] = m[i, :, :].T

        m = grid[nmain*12*n_ang:]
        mrest = m.reshape(n_ang, nrest).T
        return np.concatenate((mmain, mrest), axis=0).reshape(n_rad*n_ang)

    def _eval_nang(lmax, lv=0):
        """Summary: Lebedev order based on maximum angular momentum of given basis set
        """
        LEBEDEV_ORDER_IDX = np.array([0, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 35, 41,
                                      47, 53, 59, 65, 71, 77, 83, 89, 95, 101, 107, 113, 119, 125, 131])
        if lv == 0:   #default
            lgrids = 2*lmax+1
        elif lv == 1: #good
            lgrids = 2*lmax+7
        elif lv == 2: #verygood
            lgrids = 2*lmax+13
        elif lv == 3: #'excellent'
            lgrids = 2*lmax+20
        elif lv == 4: #'benchmark'
            lgrids = min(2*lmax+26, 131)
        minarg = np.argmin(abs(lgrids - LEBEDEV_ORDER_IDX))
        n_ang = dft.gen_grid.LEBEDEV_ORDER[LEBEDEV_ORDER_IDX[minarg]]
        return n_ang

    def eval_vhm(C, ZvH, d_atom, rad, lmax, n_coords, n_rad):
        """Summary: Python version of eval_vhm
        """

        def _eval_I1(C, rad, l, n_rad):
            I1 = np.zeros(n_rad)
            integrand = np.zeros((n_rad-1,4))
            tmp = np.zeros(n_rad-1)

            for j in range(4):
                integrand[:,j] = ( rad[1:]**(6.+l-j) - rad[:n_rad-1]**(6.+l-j) ) / (6.+l-j)

            for i in range(n_rad-2):
                tmp[i] = np.dot(integrand[i,:], C[:,i])
        
            for i in range(1,n_rad):
                I1[i] = I1[i-1] + tmp[i-1]
        
            return I1

        def _eval_I2(C, rad, l, n_rad):
            I2 =  np.zeros(n_rad)
            integrand = np.zeros((n_rad-1,4))
            tmp = np.zeros(n_rad-1)
            l = int(l)

            for j in range(4):
                if (5.-l-j) != 0 :
                    integrand[:,j] = ( rad[1:]**(5.-l-j) - rad[:n_rad-1]**(5.-l-j) ) / (5.-l-j)
                else:
                    integrand[:,6-l-1] = np.log(rad[1:]/rad[:n_rad-1])
        
            for i in range(n_rad-1):
                tmp[i] = np.dot(integrand[i,:], C[:,i])
        
            for i in range(1,n_rad):
                I2[n_rad-i-1] = I2[n_rad-i] + tmp[n_rad-i-1]

            return I2

        def _convert_coeff(c,x):
            nc=np.zeros_like(c)
            nx=np.size(x)
            for i in range(nx-1):
                nc[0,:]=c[0,:]
                nc[1,:]=c[1,:]-3*c[0,:]*x[:nx-1]
                nc[2,:]=c[2,:]-2*c[1,:]*x[:nx-1]+3*c[0,:]*x[:nx-1]**2
                nc[3,:]=c[3,:]-1*c[2,:]*x[:nx-1]+1*c[1,:]*x[:nx-1]**2-1*c[0,:]*x[:nx-1]**3
            return nc

        vh = np.zeros((ZvH.shape[0]))

        l = np.zeros((lmax+1)**2)
        for i in range(lmax+1):
            for j in range(-i, i):
                l[i*(i+1)+j] = i

        for i in range((lmax+1)**2):
            f=CubicSpline(rad,C[:,i])
            c=f.c
            c=_convert_coeff(c,rad)
            I1 = _eval_I1(c, rad, l[i], n_rad)
            I2 = _eval_I2(c, rad, l[i], n_rad)
            I = I1/(rad**(l[i]+1.)) + I2*(rad**(l[i]))
            f = CubicSpline(rad,I)
            v=f(d_atom)
            vh += ZvH[:,i]*v/(2.*l[i]+1.)
        return 4*np.pi*vh

    l_basis = np.max((mol._bas[:, 1]))
    l_margin = np.array([4, 6, 8, 12, 16])
    lmax = max(2*l_basis, l_basis + l_margin[ang_lv])

    Acoord = mol.atom_coords()
    d_atom = distance_matrix(Acoord, coords)
    symb = []
    n_rad = []
    n_ang = []
    for ia in range(mol.natm):
        symb.append(mol.atom_symbol(ia))
        chg = gto.charge(symb[ia])
        n_rad.append(dft.gen_grid._default_rad(chg, Lvl))
        n_ang.append(_eval_nang(lmax, lv=ang_lv))

    n_rad = np.max(n_rad)
    n_ang = np.max(n_ang)
    back1 = dft.gen_grid._default_rad
    back2 = dft.gen_grid._default_ang
    dft.gen_grid._default_rad = lambda *args: np.max(n_rad)
    dft.gen_grid._default_ang = lambda *args: np.max(n_ang)
    grids = dft.gen_grid.gen_atomic_grids(mol, radi_method=radi_method, prune=None, level=Lvl)
    dft.gen_grid._default_rad = back1
    dft.gen_grid._default_ang = back2

    rad, dr = radi_method(n_rad, chg)
    wrad = rad**2.*dr  #Radial weights
    sample_r = int(n_rad/2)
    ng = (grids[symb[0]][1]).size
    c = np.zeros((mol.natm, ng, 3))
    r = np.zeros((mol.natm, ng))
    w = np.zeros((mol.natm, ng))
    wang = np.zeros((mol.natm, n_ang))
    ao = np.zeros((mol.natm, ng, mol.nao_nr()), order='F')
    p = np.zeros((mol.natm, mol.natm, ng))
    ZI = np.zeros((mol.natm, n_ang, (lmax+1)**2), order='F')
    ZvH = np.zeros((mol.natm, coords.shape[0], (lmax+1)**2))

    #Density independent values
    for j, ia in enumerate(symb):    #j : idx , ia : name
        ca = np.array(grids[symb[j]][0]) #xyz coordinate centred at symb atom
        cx = _grid_refine(n_rad, n_ang, ca[:, 0])
        cy = _grid_refine(n_rad, n_ang, ca[:, 1])
        cz = _grid_refine(n_rad, n_ang, ca[:, 2])
        c[j] = np.vstack((cx, cy, cz)).T
        r[j] = np.linalg.norm(c[j], axis=1)
        ZI2 = _Cart_Spharm(c[j][sample_r*n_ang:(sample_r+1)*n_ang], lmax)
        wa = np.array(grids[symb[j]][1]) #weights
        w[j] = _grid_refine(n_rad, n_ang, wa)
        wang = w[j][sample_r*n_ang:(sample_r+1)*n_ang]/wrad[sample_r] #Angular weights
        ZI[j] = np.einsum('al,a->al', ZI2, wang)
        tst = c[j]+Acoord[j]
        d = distance_matrix(Acoord, tst) #the difference between newly define grids and original grids
        rel_coord = coords-Acoord[j, :]
        """.. todo::
            * Multi threading on "ZvH[j]=_Cart_Spharm(rel_coord, lmax)"
        """
        ZvH[j] = _Cart_Spharm(rel_coord, lmax) #This is time consuming
        #partition function P_i
        p[j] = np.exp(-2.*d)/(d**3) #partition function P_i
        for ia, za in enumerate(mol.atom_charges()):
            if za==1: #Special treatment on hydrogen atom
                p[j, ia, :] *= 0.3
        ao[j] = numint.eval_ao(mol, tst) #AO value in real coordinate

    #Density dependent values
    vH = np.zeros(int(coords.size/3))
    for i in range(mol.natm):
        idx = np.argsort(d_atom[i])
        rho_org = numint.eval_rho(mol, ao[i], dm) #Rho in real coordinate
        rho_here = p[i, i, :] / np.sum(p[i], axis=0)*rho_org #Eq(4)
        rho_here = rho_here.reshape(n_rad, n_ang) #(r,\theta \phi)
        #r : n_rad, a : n_ang, l : (lmax+1)**2
        C = np.matmul(rho_here, ZI[i])
        if kf_imported:
            vH[idx] += kspies_fort.eval_vhm(C, ZvH[i, idx, :], d_atom[i, idx], rad, lmax, coords.shape[0], n_rad)
        else:
            vH[idx] += eval_vhm(C, ZvH[i, idx, :], d_atom[i, idx], rad, lmax, coords.shape[0], n_rad)

    return vH

def eval_vxc(mol, dm, xc_code, coords, delta=1e-7):
    """Summary: Calculate real-space exchange-correlation potential for GGA from given density matrix

        Args:
            mol (object) : an instance of :class:`Mole`
            coords (ndarray) : grids space used for calculating XC potential
            dm (ndarray) : one-particle reduced density matrix in basis set representation
            xc_code (str) : XC functional description.
            delta (float) : amount of finite difference to calculate numerical differentiation, default is 1e-7 a.u.

        Returns:
            (tuple): tuple containing:

                (ndarray) Pointwise XC potential, vxc(r) for RKS dm, vxc_alpha(r) for UKS dm

                (ndarray) Pointwise XC potential, vxc(r) for RKS dm, vxc_beta(r) for UKS dm


    """
    Ncoord = np.size(coords)//3
    ao = numint.eval_ao(mol, coords, deriv=1)

    def _spin_tr(dm):
        """Summary: Return spin-traced density matrix
        """
        if np.array(dm).ndim == 3:
            return dm[0]+dm[1]
        return dm

    def _numderiv(aux, delta):
        """Summary: numerical differentiation of 3D function
        """
        nabla_res = np.zeros((3, Ncoord))
        nabla_res[0, :] = (aux[0, :]-aux[1, :])/(2.*delta)
        nabla_res[1, :] = (aux[2, :]-aux[3, :])/(2.*delta)
        nabla_res[2, :] = (aux[4, :]-aux[5, :])/(2.*delta)
        return nabla_res

    auxcoords = np.zeros((6, Ncoord, 3))
    for i in range(6):
        auxcoords[i, :, :] = coords[:, :]
        auxcoords[i, :, (i//2)] = coords[:, (i//2)]+delta*(-1.)**i

    if mol.spin == 0: #spin-unpolarized case
        dm = _spin_tr(dm)

        den = numint.eval_rho(mol, ao, dm, xctype='GGA')
        exc, vxc = dft.libxc.eval_xc(xc_code, den, spin=mol.spin, deriv=1)[:2]

        auxvsigma = np.zeros((6, Ncoord))
        for i in range(6):
            auxao = numint.eval_ao(mol, auxcoords[i, :, :], deriv=1)
            auxden = numint.eval_rho(mol, auxao, dm, xctype='GGA')
            vxc = dft.libxc.eval_xc(xc_code, auxden, spin=mol.spin, deriv=1)[1]
            auxvsigma[i, :] = vxc[1]

        ao = numint.eval_ao(mol, coords, deriv=2)
        den = numint.eval_rho(mol, ao, dm, xctype='mGGA')
        nabla_vsigma = _numderiv(auxvsigma, delta)
        vxc = vxc[0]-2*(den[4, :]*vxc[1]+np.einsum('ir,ir->r', den[1:4, :], nabla_vsigma[:, :]))
        if np.array(dm).ndim == 2: #RKS scheme
            return np.array(vxc)
        elif np.array(dm).ndim == 3: #UKS scheme
            return np.array(vxc), np.array(vxc)

    elif mol.spin != 0: #spin-polarized case
        den_a = numint.eval_rho(mol, ao, dm[0], xctype='GGA')
        den_b = numint.eval_rho(mol, ao, dm[1], xctype='GGA')
        exc, vxc = dft.libxc.eval_xc(xc_code, (den_a, den_b), spin=mol.spin, deriv=1)[:2]

        auxvsigma_aa = np.zeros((6, Ncoord))
        auxvsigma_ab = np.zeros((6, Ncoord))
        auxvsigma_bb = np.zeros((6, Ncoord))
        for i in range(6):
            auxao = numint.eval_ao(mol, auxcoords[i, :, :], deriv=1)
            auxden_a = numint.eval_rho(mol, auxao, dm[0], xctype='GGA')
            auxden_b = numint.eval_rho(mol, auxao, dm[1], xctype='GGA')
            vxc = dft.libxc.eval_xc(xc_code, (auxden_a, auxden_b), spin=mol.spin, deriv=1)[1]
            auxvsigma_aa[i, :] = vxc[1][:, 0]
            auxvsigma_ab[i, :] = vxc[1][:, 1]
            auxvsigma_bb[i, :] = vxc[1][:, 2]

        nabla_vsigma_aa = _numderiv(auxvsigma_aa, delta)
        nabla_vsigma_ab = _numderiv(auxvsigma_ab, delta)
        nabla_vsigma_bb = _numderiv(auxvsigma_bb, delta)

        ao = numint.eval_ao(mol, coords, deriv=2)
        den_a = numint.eval_rho(mol, ao, dm[0], xctype='mGGA')
        den_b = numint.eval_rho(mol, ao, dm[1], xctype='mGGA')

        vxc_a = vxc[0][:, 0]\
        -2.*(den_a[4, :]*vxc[1][:, 0]+np.einsum('ir,ir->r', den_a[1:4, :], nabla_vsigma_aa[:, :]))\
        -np.einsum('ir,ir->r', nabla_vsigma_ab, den_b[1:4, :])-den_b[4, :]*vxc[1][:, 1]

        vxc_b = vxc[0][:, 1]\
        -2.*(den_b[4, :]*vxc[1][:, 2]+np.einsum('ir,ir->r', den_b[1:4, :], nabla_vsigma_bb[:, :]))\
        -np.einsum('ir,ir->r', nabla_vsigma_ab, den_a[1:4, :])-den_a[4, :]*vxc[1][:, 1]

        return np.array(vxc_a), np.array(vxc_b)



"""
Zhao-Morrison-Parr
==================

**Summary** This script preforms a Zhao-Morrison-Parr [ZMP1994]_ Kohn Sham inversion.

Written for Python 3.7.4

  :References:

    .. [ZMP1994] Qingsheng Zhao, Robert C Morrison, and Robert G Parr.
        From electron densities to Kohn-Sham kinetic energies, orbital energies, exchange-correlation potentials, and exchange-correlation energies. (1994)
        <https://doi.org/10.1103/PhysRevA.50.2138> Physical Review A, 50(3) 2138.

    .. [THG1997] David J Tozer, Nicholas C Handy, and William H Green.
        Exchange-correlation functionals from ab initio electron densities (1997)
        <https://doi.org/10.1016/S0009-2614(97)00586 and-1> Chemical Physics Letters, 273(3-4) 183-194

    .. [Pulay1980] P Pulay.
        Convergence acceleration of iterative sequences. the case of SCF iteration (1980)
        <https://doi.org/10.1016/0009-2614(80)80396-4> Chemical Physics Letters, 73 (2): 393–398

.. moduleauthor::
    Seungsoo Nam <skaclitz@yonsei.ac.kr> <http://tccl.yonsei.ac.kr/mediawiki/index.php/Main_Page> ORCID: `000-0001-9948-6140 <https://orcid.org/0000-0001-9948-6140>`_

.. topic:: Funding

    This research was made possible by funding from the National Research Foundation of Korea (NRF-2020R1A2C2007468 and NRF-2020R1A4A1017737).

.. topic:: Internal Log

    **2020-06-13** RJM added doc string templates

    **2020-08-21** SN corrected typos, minor changes in attribute names, etc.

"""

from functools import reduce
import numpy as np
from kspies import util
from pyscf import scf, dft


class DIIS:
    """Summary: Class for DIIS extrapolation used in ZMP

       See [Pulay1980]_ for some extra context.

    """
    def __init__(self, S, diis_space):
        """Initialize DIIS object

        Args:
            S (ndarray):  overlap integral
            diis_space (integer) : number of DIIS vectors used in extrapolation

        """
        eig, Z = np.linalg.eigh(S)
        S12 = 1./np.sqrt(eig)
        self.S = S
        self.O = reduce(np.dot, (Z, np.diag(S12), Z.T))
        self.diis_space = diis_space
        self.norb = len(S[0])
        self.ems = np.zeros((self.diis_space, self.norb, self.norb))
        self.pms = np.zeros((self.diis_space, self.norb, self.norb))
        self.tall = self.t_1 = self.t_2 = self.t_3 = 0.

    def extrapolate(self, iteration, fock, dm):
        """Summary: New fock matrix by linear combination of previous fock matrices

        Args:
            iteration (integer) : present SCF iteration
            fock (ndarray) : fock matrix
            dm (ndarray) : density matrix obtained from previous step

        Returns:
            (tuple): tuple containing:

                (ndarray): **newfock** extrapolated fock matrix

                (float): **diis_error** DIIS error used in convergence criteria

        """

        if iteration <= 1 or self.diis_space < 2:
            return fock, 0.0

        for k in range(1, min(iteration, self.diis_space))[::-1]:
            self.ems[k] = self.ems[k-1]
            self.pms[k] = self.pms[k-1]

        em = reduce(np.dot, (fock, dm, self.S))
        em -= em.T
        self.ems[0] = reduce(np.dot, (self.O.T, em, self.O))
        self.pms[0] = fock[:]
        idx = np.abs(self.ems[0]).argmax()
        diis_error = np.abs(np.ravel(self.ems[0])[idx])

        #Solve BC = A to find C
        nb = min(iteration, self.diis_space)-1
        B = -1.*np.ones((nb+1, nb+1))
        B[nb, nb] = 0.
        B[:nb, :nb] = np.einsum('aij,bji->ab', self.ems[:nb, :, :], self.ems[:nb, :, :],optimize='greedy')
        A = np.zeros(nb+1)
        A[nb] = -1.
        C = np.linalg.solve(B, A)

        # form new extrapolated diis fock matrix
        newfock = np.zeros_like(fock)
        for i, c in enumerate(C[:-1]):
            newfock += c*self.pms[i]

        return newfock, diis_error

def basic(mz, mol):
    """Summary: Common basic initialization function for RZMP and UZMP objects

    Args:
        mz : RZMP or UZMP object
        mol (object) : an instance of :class:`Mole`

    """
    mz.mol = mol
    mz.guide = 'faxc'
    mz.diis_space = 40
    mz.level_shift = .2
    mz.max_cycle = 400
    mz.conv_tol_dm = 1e-7
    mz.conv_tol_diis = 1e-5
    mz.with_df = False
    mz.verbose = mz.mol.verbose
    mz.stdout = mz.mol.stdout

    mz.S = mz.mol.intor_symmetric('int1e_ovlp')
    mz.T = mz.mol.intor_symmetric('int1e_kin')
    mz.V = mz.mol.intor_symmetric('int1e_nuc')

class RZMP:
    """Summary: Perform ZMP calculation in restricted scheme, see [ZMP1994]_ for detail.

    .. _restricted-zmp:

    Attributes:
        mol (object) : an instance of :class:`Mole`
        dm_tar (ndarray) : Density matrix of target density in atomic orbital basis representation
        dm_aux (ndarray) : Auxilary density matrix to construct a fixed part of fock matrix. Default is dm_tar
        guide (str) : Guiding potential. Can be set as

            |  None : no guiding potential except external potential
            |  'faxc' : Exchange-correlation part of Fermi-Amaldi potential
            |  xc   : ks.xc attribute in pyscf DFT

        diis_space (int) : DIIS space size. Default is 40
        level_shift (float) : Level shift (in AU) for virtual space. Default is 0.2
        max_cycle (int) : max number of zscf iterations. Defalut is 400
        conv_tol_dm (float) :  converge threshold for density matrix. Default is 1e-7
        conv_tol_diis (float) : converge threshold for DIIS error. Default is 1e-5

    :ivar:
        * **converged** (bool) – zscf converged or not
        * **mo_energy** (ndarray) –  molecular orbital energies (Note that energies when level_shift = 0)
        * **mo_occ** (ndarray) –  molecular orbital occupation numbers
        * **mo_coeff** (ndarray) – molecular orbital coefficients
        * **dm** (ndarray) – density matrix in atomic orbital basis representation
        * **l** (float) – the last given lambda

    """
    def __init__(self, mol, dm_tar, dm_aux=None):
        basic(self,mol)

        self.dm_tar = dm_tar
        self.initialized = False
        self.dm = dm_tar
        self.dm_aux = dm_aux
        self.dm_old = dm_tar
        self.verbose = mol.verbose

    get_occ = scf.hf.get_occ
    make_rdm1 = scf.hf.SCF.make_rdm1

    def initialize(self):
        """Summary: Construct a fixed part of fock matrix F0

        .. math::

            F_{0} = T + V_{ext} + V_{h} + V_{g}

        """
        self.get_ovlp=lambda *args: self.S

        if self.with_df:
            self.mf=dft.RKS(self.mol).density_fit()
        else:
            self.mf=dft.RKS(self.mol)
        self.mf.grids.build()
        self.coords=self.mf.grids.coords
        self.weights=self.mf.grids.weights
        self.ao=dft.numint.eval_ao(self.mol,self.coords)

        if self.dm_aux is None :
            self.dm_aux=self.dm_tar
        self.J_tar=self.mf.get_jk(self.mol,self.dm_tar)[0]

        if self.guide is None:
            self.V0=np.zeros_like(self.dm_tar)
        else:
            fac_faxc, dft_xc = util.parse_guide(self.guide)

            N = self.mol.nelectron
            J_tar = scf.hf.get_jk(self.mol, self.dm_aux)[0]
            VFA = -(1./N)*(J_tar)

            self.mf.xc = dft_xc
            Vxcdft = self.mf.get_veff(self.mol, dm=self.dm_aux)

            self.V0 = fac_faxc * VFA + Vxcdft

        self.F0=self.T+self.V+self.V0
        self.initialized=True

    def zscf(self, l):
        """Summary:
            Run self-consistent ZMP equation under given lambda (l). Prints lambda,
            HOMO-LUMO, dN and C to terminal.

            Args:
                l (float): Lagrange multiplier lambda
        """
        if not self.initialized:
            self.initialize()
        self.l = l

        self.converged = False
        self.zdiis = DIIS(self.S, self.diis_space)

        for cycle in range(1, self.max_cycle):
            self.J = self.mf.get_jk(self.mol, self.dm)[0]

            self.F = self.F0 + l*(self.J - self.J_tar)
            self.F = scf.hf.level_shift(self.S, self.dm*.5, self.F, self.level_shift)
            self.F, diis_e = self.zdiis.extrapolate(cycle, self.F, self.dm) #DIIS

            self.mo_energy, self.mo_coeff = scf.hf.eig(self.F, self.S)
            self.mo_occ = self.get_occ(self.mo_energy, self.mo_coeff)
            self.dm = self.make_rdm1(self.mo_coeff, self.mo_occ)

            ddm = self.dm_old - self.dm
            dm_e = np.max(np.abs(ddm))
            self.dm_old = self.dm
            dm_converged = dm_e < self.conv_tol_dm
            diis_converged = diis_e < self.conv_tol_diis
            self.mo_energy[self.mo_occ==0] -= self.level_shift

            nocc = self.mol.nelectron // 2
            HOMO, LUMO =self.mo_energy[nocc-1], self.mo_energy[nocc]
            gap = LUMO - HOMO

            print(f'\rlambda= {l:7.2f}  iter: {cycle:4d} gap= {gap:10.7f}   ',end='\r')

            self.converged = dm_converged and diis_converged
            if self.converged and cycle > 1:
                break

        self.J = self.mf.get_jk(self.mol, self.dm)[0]
        dn = dft.numint.eval_rho(self.mol, self.ao, self.dm - self.dm_tar)
        dN = 1000*np.einsum('r,r', abs(dn), self.weights)
        C = np.einsum('ij,ji', self.dm-self.dm_tar, self.J - self.J_tar) #Objective of minimization
        print(f'lambda= {l:7.2f} niter: {cycle:4d} gap= {LUMO-HOMO:10.7f} dN= {dN:7.2f} C= {C:.2e} ')

class UZMP:
    """Summary: Perform ZMP calculation in unrestricted scheme, see [THG1997]_.

    .. _unrestricted-zmp:

    Attributes:
        mol (object) : an instance of :class:`Mole`
        dm_tar (ndarray) : Density matrix of target density in atomic orbital basis representation
        dm_aux (ndarray) : Auxilary density matrix to construct a fixed part of fock matrix. Default is dm_tar
        guide (str) : Guiding potential. Can be set as

            |  None : no guiding potential except external potential
            | 'faxc' : Exchange-correlation part of Fermi-Amaldi potential
            |  xc   : ks.xc attribute in pyscf DFT

        diis_space (int) : DIIS space size. Default is 40
        level_shift (float) : Level shift (in AU) for virtual space. Default is 0.2
        max_cycle (int) : max number of zscf iterations. Defalut is 400
        conv_tol_dm (float) :  converge threshold for density matrix. Default is 1e-7
        conv_tol_diis (float) : converge threshold for DIIS error. Default is 1e-5

    :ivar:
        * **converged** (bool) – zscf converged or not
        * **mo_energy** (ndarray) –  molecular orbital energies (Note that energies when level_shift = 0)
        * **mo_occ** (ndarray) –  molecular orbital occupation numbers
        * **mo_coeff** (ndarray) – molecular orbital coefficients
        * **dm** (ndarray) – density matrix in atomic orbital basis representation
        * **l** (float) – the last given lambda

    """
    def __init__(self, mol, dm_tar, dm_aux=None):
        basic(self, mol)
        self.nelec = [int((mol.nelectron + mol.spin)//2), int((mol.nelectron - mol.spin)//2)]

        self.dm_tar = dm_tar
        self.initialized = False
        self.dm = dm_tar
        self.dm_aux = dm_aux
        self.dm_old = dm_tar
        self.verbose = mol.verbose

    get_occ = scf.uhf.get_occ
    make_rdm1 = scf.uhf.UHF.make_rdm1
    spin_square = scf.uhf.UHF.spin_square

    def initialize(self):
        """Summary: Construct a fixed part of fock matrix F0

        .. math::

            F_{0} = T + V_{ext} + V_{h} + V_{g}

        """
        self.get_ovlp=lambda *args: self.S

        if self.with_df:
            self.mf = dft.UKS(self.mol).density_fit()
        else:
            self.mf = dft.UKS(self.mol)
        self.mf.grids.build()
        self.coords = self.mf.grids.coords
        self.weights = self.mf.grids.weights
        self.ao = dft.numint.eval_ao(self.mol,self.coords)

        if self.dm_aux is None:
            self.dm_aux = self.dm_tar
        self.J_tar = self.mf.get_jk(self.mol,self.dm_tar)[0]

        if self.guide is None:
            self.V0 = np.zeros_like(self.dm_tar)

        else:
            fac_faxc, dft_xc = util.parse_guide(self.guide)

            N = self.mol.nelectron
            J_tar = scf.hf.get_jk(self.mol, self.dm_aux)[0]
            VFA = -(1./N)*(J_tar[0]+J_tar[1])

            self.mf.xc = dft_xc
            Vxcdft = self.mf.get_veff(self.mol, dm=self.dm_aux)

            self.V0 = fac_faxc * np.array((VFA, VFA)) + Vxcdft

        self.F0 = (self.T + self.V+self.V0[0],
                   self.T + self.V+self.V0[1])
        self.initialized = True

    def zscf(self, l):
        """Summary: Run self-consistent ZMP equation under given lambda (l)

            Args:
            l (float): Lagrange multiplier lambda

        """
        if not self.initialized:
            self.initialize()
        self.l = l

        self.converged = False
        self.zdiis_a = DIIS(self.S, self.diis_space)
        self.zdiis_b = DIIS(self.S, self.diis_space)

        for cycle in range(1, self.max_cycle):
            self.J = self.mf.get_jk(self.mol, self.dm)[0]

            self.Fa = self.F0[0] + 2*l*(self.J[0] - self.J_tar[0])
            self.Fb = self.F0[1] + 2*l*(self.J[1] - self.J_tar[1])

            self.Fa = scf.hf.level_shift(self.S, self.dm[0], self.Fa, self.level_shift)
            self.Fb = scf.hf.level_shift(self.S, self.dm[1], self.Fb, self.level_shift)

            self.Fa, diis_e_a = self.zdiis_a.extrapolate(cycle, self.Fa, self.dm[0])
            self.Fb, diis_e_b = self.zdiis_b.extrapolate(cycle, self.Fb, self.dm[1])

            e_a, c_a = scf.hf.eig(self.Fa, self.S)
            e_b, c_b = scf.hf.eig(self.Fb, self.S)
            self.mo_energy = np.array((e_a,e_b))
            self.mo_coeff = np.array((c_a,c_b))

            self.mo_occ = self.get_occ(self.mo_energy, self.mo_coeff)
            self.dm = self.make_rdm1(self.mo_coeff, self.mo_occ)

            ddm = self.dm_old - self.dm
            dm_e = np.max(np.abs(ddm))
            self.dm_old = self.dm
            dm_converged = dm_e < self.conv_tol_dm
            diis_converged = diis_e_a+diis_e_b < self.conv_tol_diis
            self.mo_energy[0][self.mo_occ[0]==0] -= self.level_shift
            self.mo_energy[1][self.mo_occ[1]==0] -= self.level_shift

            HOMO = np.maximum(self.mo_energy[0][self.nelec[0]-1],
                              self.mo_energy[1][self.nelec[1]-1])
            LUMO = np.minimum(self.mo_energy[0][self.nelec[0]],
                              self.mo_energy[1][self.nelec[1]])
            gap = LUMO-HOMO

            print(f'\rlambda= {l:7.2f}  iter: {cycle:4d} gap= {gap:10.7f}   ',end='\r')

            self.converged = dm_converged and diis_converged
            if self.converged and cycle > 1:
                break

        self.J = self.mf.get_jk(self.mol, self.dm)[0]
        #Calculate alpha/beta density difference separately
        #dn_a = dft.numint.eval_rho(self.mol, self.ao, (self.dm-self.dm_tar)[0])
        #dn_b = dft.numint.eval_rho(self.mol, self.ao, (self.dm-self.dm_tar)[1])
        #dN = 1000*np.einsum('r,r', abs(dn_a)+abs(dn_b), self.weights)
        dn = dft.numint.eval_rho(self.mol, self.ao, (self.dm[0]+self.dm[1]-self.dm_tar[0]-self.dm_tar[1]))
        dN = 1000*np.einsum('r,r', abs(dn), self.weights)

        Ca = np.einsum('ij,ji', self.dm[0]-self.dm_tar[0], self.J[0]-self.J_tar[0])
        Cb = np.einsum('ij,ji', self.dm[1]-self.dm_tar[1], self.J[1]-self.J_tar[1])
        C = 2*(Ca+Cb)
        print(f'lambda= {l:7.2f} niter: {cycle:4d} gap= {gap:10.7f} dN= {dN:7.2f} C= {C:.2e}  ')




"""
Wu and Yang
===========

**Summary** This is a program for preforming Kohn-Sham inversion with algorithm introduced by Wu and Yang [WuYang2003]_.

Written for Python 3.7.4 Lots of other text here, like details about how this uses the original Wu Yang idea [WuYang2003]_, and how the other ref is used for  Regularization [HBY2007]_ . Any useful info about the algorithm should also be given.

.. moduleauthor::
    Seungsoo Nam <skaclitz@yonsei.ac.kr> <http://tccl.yonsei.ac.kr/mediawiki/index.php/Main_Page> ORCID: `000-0001-9948-6140 <https://orcid.org/0000-0001-9948-6140>`_
    Ryan J. McCarty <rmccarty@uci.edu> <http://ryanjmccarty.com> ORCID: `0000-0002-2417-8917 <http://https://orcid.org/0000-0002-2417-8917>`_

:References:

    .. [WuYang2003] Qin Wu and Weitao Yang. A direct optimization method for calculating density functionals and exchange–correlation potentials from electron densities (2003)
        <https://doi.org/10.1063/1.1535422> Journal of Chemical Physics, 118(6).

    .. [HBY2007] Tim Heaton-Burgess, Felipe A. Bulat, and Weitao Yang. Optimized Effective Potentials in Finite Basis Sets (2007)
        <https://doi.org/10.1103/PhysRevLett.98.256401> Physical review letters, 98(25).

    .. [GNGK2020] Rachel Garrick, Amir Natan, Tim Gould, and Leeor Kronik. Exact Generalized Kohn-Sham Theory for Hybrid Functionals (2020)
        <https://doi.org/10.1103/PhysRevX.10.021040> Physical Review X, 10(2).

.. topic:: Funding

    This research was made possible by funding from the National Research Foundation of Korea (NRF-2020R1A2C2007468 and NRF-2020R1A4A1017737).

.. topic:: Internal Log

    **2020-06-02** RJM added in docstring templates

    **2020-06-15** SN added some comments and fix bug(RWY and UWY gave different results with regularization)

    **2020-08-21** SN corrected typos, minor changes in attribute names, etc.

    **2020-11-12** guiding potential parser

    **2021-01-07** Analytical three-center overlap integral with auxiliary basis

    **2021-02-08** Support target density matrix in different basis representation

"""

import time, warnings
import numpy as np
from scipy.optimize import minimize
from pyscf import gto, scf, dft, df
from kspies import util
from opt_einsum import contract
try:
    from kspies import kspies_fort
    kf_imported=True
except:
    kf_imported=False
#kf_imported=False

def project_b(mw1, mw2):
    """Summary: Transfer b vector from one WY calculation to other
    """
    mol1 = mw1.pmol
    npbs1 = mol1.nao_nr()
    b1 = mw1.b
    mol2 = mw2.pmol
    if type(mw1).__name__ == 'RWY' and type(mw2).__name__ == 'RWY':
        b1 = np.array([b1]).T
        mw2.b = scf.addons.project_mo_nr2nr(mol1, b1, mol2)[:,0]
    elif type(mw1).__name__ == 'UWY' and type(mw2).__name__ == 'UWY':
        b1 = np.vstack((b1[:npbs1],b1[npbs1:])).T
        b2 = scf.addons.project_mo_nr2nr(mol1, b1, mol2)
        mw2.b = np.hstack((b2[:,0],b2[:,1]))
    else:
        warnings.warn("Two WY objects are not same!")

def numint_3c2b(mol, pbas, level=5):
    """Summary: Three-center overlap integral with different atomic- and potential- basis sets with numerical integration

    Args:
        mol : an instance of :class:`Mole`
        pbas (list) : potential basis set for WY, given as a list of functions 
        level (int 0~9) : PySCF preset mesh grid used for numerical integration. Default is 5

    Returns:
        (ndarray): **Sijt** three-center overlap integral with shape ((n1, n1, n2))
        where n1 is the length of atomic orbital basis set defined in mol,
        and n2 is the length of potential basis set

    """
    grids = dft.gen_grid.Grids(mol)
    grids.level = level
    grids.build()
    coords = grids.coords
    weights = grids.weights
    ao1 = dft.numint.eval_ao(mol, coords, deriv=0)
    ao2 = []
    n1 = mol.nao_nr()
    n2 = 0

    for vr_generator in pbas:
       n2 += 1
       ao2.append(vr_generator(coords))
    ao2=np.array(ao2).T

    t0 = time.time()
    if kf_imported: #Utilize faster numerical integration
        Sijt = kspies_fort.ovlp_aab(weights, ao1, ao2, n1, n2, len(weights))
    else: #kspies_fort not imported, us numpy
        Sijt = contract('ri, rj, rk, r->ijk', ao1, ao1, ao2, weights)
    n1 = mol.nao_nr()
    t = (time.time()-t0)/60.
    print("Three-center overlap integral by numerical integration")
    print("n1  : %4i"%n1)
    print("n2  : %4i"%n2)
    print("time: %.1f min"%t)
    return Sijt

def time_profile(mw):
    """Summary: Print to terminal a collection of time data in seconds

    Terminal prints:
    * initialize : time for initialization
    * total_opt  : total elasped time
    * solve_eig  : time to construct fock matrix and diagonalize it
    * Ws_eval    : time to evaluate objective function Ws
    * Gd_eval    : time to evaluate gradient of objective function
    * Hs_eval    : time to evaluate hessian of objective function

    """
    print("*********Time Profile*********")
    print("initialize : %15.3f"%mw.t_init)
    print("total_opt  : %15.3f"%mw.t_opt)
    print("solve_eig  : %15.3f"%mw.t_eig)
    print("Ws_eval    : %15.3f"%mw.t_ws)
    print("Gd_eval    : %15.3f"%mw.t_gd)
    print("Hs_eval    : %15.3f"%mw.t_hs)

def run(mw, xbas=None):
    """Summary: Minimize an objective function -Ws

    Args:
        mw : RWY or UWY object

    """
    mw.initialize(xbas)
    t = time.time()
    if mw.method.lower()=='cg' or mw.method.lower()=='bfgs':
        mw.res = minimize(mw.eval_Ws, x0=mw.b, method=mw.method, jac=mw.eval_Gd,
                          tol=mw.tol, options={'disp': False})
    else:
        mw.res = minimize(mw.eval_Ws, x0=mw.b, method=mw.method, jac=mw.eval_Gd, hess=mw.eval_Hs,
                          tol=mw.tol, options={'disp': False})
    mw.b = mw.res.x
    mw.converged = mw.res.success
    mw.t_opt += time.time()-t

def wyscf(mw, ddmtol=1e-6):
    """Summary: Run WY until its self-consistent. This is required for hybrid guiding potentials

    .. note::
        This function performs very simplied generalization of WY for non-local guiding potential, mimicking [GNGK2020]_.
        There is no theoretical proof that the result of this function can be trusted.
        The optimization of WY objective functional is only theoretically proven for local multiplicative potentials, not non-local potential.

    Args:
        mw : RWY or UWY object
        ddmtol (float, optional) : Convergence criteria of density matrix.

    """
    mw.totnit = 0
    mw.run()
    mw.totnit += mw.res.nit
    mw.scfcycle = 0

    for i in range(40):
        mw.scfcycle = i+1
        dm_old = mw.dm
        mw.dm_aux = dm_old
        mw.run(xbas=mw.mol.basis)
        mw.totnit += mw.res.nit
        mw.ddm = np.max(np.abs(np.array(dm_old)-np.array(mw.dm)))
        if mw.ddm < ddmtol:
            break

def info(mw):
    """Summary: Print summary of minimization

    Args:
        mw : RWY or UWY object

    """
    grd = mw.res.jac
    maxgrd = np.max(abs(grd))
    if mw.converged == False:
        print(" *****Optimization Failed*****")
        print("      after %i iterations "%mw.res.nit)
    else:
        print("****Optimization Completed****")
        if hasattr(mw, 'scfcycle'):
            print("      after %i scf cycles "%mw.scfcycle)
            print("      and %i iterations "%mw.totnit)
            print("max(ddm)   : %15.8f"%mw.ddm)
        else:
            print("      after %i iterations "%mw.res.nit)
    print("func_value : %15.8f"%mw.res.fun)
    print("max_grad   : %15.8f"%maxgrd)

def basic(mw, mol, pbas, Sijt, tbas, Smnt):
    """Summary: Common basic initialization function for RWY and UWY objects

    Args:
        mw : RWY or UWY object
        mol : an instance of :class:`Mole`
        pbas (dict or str) : potential basis set for WY
        Sijt (ndarray) : three-center overlap integral with shape ((no, no, np))
          where no is the length of atomic orbital basis set defined in mol,
          while np is the length of potential basis set
          Only effective for model systems and molecular systems with user-defined pbs.
          Otherwise, this is ignored and analytical integration is performed.
        tbas (dict or str) : basis set of target density matrix
        Smnt (ndarray) : three-center overlap integral with shape ((nt, nt, np))
          where nt is the length of atomic orbital basis set that defines target density matrix,
          while np is the length of potential basis set
          Only effective for model systems and molecular systems with user-defined pbs.
          Otherwise, this is ignored and analytical integration is performed.
    """

    mw.mol = mol
    mw.guide = 'faxc'
    mw.reg = 0.
    mw.tol = 1e-6
    mw.method = 'trust-exact'
    mw.verbose = mw.mol.verbose
    mw.stdout = mw.mol.stdout

    mw.Sijt = None
    mw.Smnt = None
    mw.Tp = None
    mw.pbas = pbas
    mw.tbas = tbas

    is_model1 = len(mw.mol._atm) == 0
    is_model2 = len(mw.mol._bas) == 0
    mw.model = is_model1 and is_model2
    if mw.model: #Check if defined mol is a molecule or model
        mw.guide = None
        if Sijt is None:
            raise AssertionError("Three-center overlap integeral should be given for model system")
        if kf_imported:
            mw.Sijt = np.array(Sijt, order='F')
            if Smnt is not None:
                mw.Smnt = np.array(Smnt, order='F')
        else:
            mw.Sijt = np.array(Sijt, order='C')
            if Smnt is not None:
                mw.Smnt = np.array(Smnt, order='C')
        return

    mw.S = mw.mol.intor_symmetric('int1e_ovlp')
    mw.T = mw.mol.intor_symmetric('int1e_kin')
    mw.V = mw.mol.intor_symmetric('int1e_nuc')

    if tbas is None:
        mw.tbas = mol.basis
    mw.tmol = df.make_auxmol(mol, auxbasis=mw.tbas)

    if pbas is not None and callable(pbas[0]):
        #potential basis is given as a list of user-defined functions
        #In this case, mw.Tp should be given by the user to use regularization
        mw.Sijt = Sijt
        if Sijt is None:
            mw.Sijt = numint_3c2b(mol, pbas)

        if not gto.mole.same_basis_set(mol, mw.tmol):
            mw.Smnt = Smnt
            if Smnt is None:
                mw.Smnt = numint_3c2b(mw.tmol, pbas)

    else:
        #potential basis is given as a pyscf-supported format
        if pbas is None:
            mw.pbas = mol.basis

        mw.pmol = df.make_auxmol(mol, auxbasis=mw.pbas)

        mw.Tp = mw.pmol.intor_symmetric('int1e_kin')
        mw.Sijt = df.incore.aux_e2(mol, mw.pmol, intor='int3c1e')

        if not gto.mole.same_basis_set(mol, mw.tmol):
            mw.Smnt = df.incore.aux_e2(mw.tmol, mw.pmol, intor='int3c1e')

    mw.nbas = len(mw.Sijt[:, 0, 0])
    mw.npot = len(mw.Sijt[0, 0, :])


class RWY:
    """Summary: Perform WY calculation in restricted scheme

    .. _restrictedwy:

    Attributes:
        mol (object) : an instance of :class:`Mole`
        dm_tar (ndarray) : Density matrix of target density in atomic orbital basis representation
        pbas (dict or str or list) : Potential basis set for WY. If not given, same with atomic orbital basis. 
                                     If this is given as a list of function, those functions are recognized as potential basis set.
        Sijt (ndarray) : Three-center overlap integral. Ignored (calculated analitically) when pbas is given as dict or str.
        dm_aux (ndarray) : Auxilary density matrix to construct density-dependent part of guiding potential. Default is dm_tar
        guide (str) : Guiding potential. Can be set as

            |  None : no guiding potential except external potential
            |  'faxc' : Exchange-correlation part of Fermi-Amaldi potential
            |  xc   : ks.xc attribute in pyscf DFT

        reg (float) : strength of regularization. Default is 0
        tol (float) : tolarance of optimization (maximum gradient element). Default is 1e-6
        method (str) : optimization algorithm used in SciPy. Can be (case insensitive) 'cg', 'bfgs', 'newton-cg',
                       'dogleg', 'trust-ncg', 'trust-krylov', 'trust-exact'. Defult is 'trust-exact'

    :ivar:
        * **converged** (bool) – optimization converged or not
        * **mo_energy** (ndarray) –  molecular orbital energies
        * **mo_occ** (ndarray) –  molecular orbital occupation numbers
        * **mo_coeff** (ndarray) – molecular orbital coefficients
        * **dm** (ndarray) – density matrix in atomic orbital basis representation
        * **b** (ndarray) – optimized linear combination coefficients

    """
    def __init__(self, mol, dm_tar, pbas=None, Sijt=None, tbas=None, Smnt=None, dm_aux=None):
        t = time.time()
        basic(self, mol, pbas, Sijt, tbas, Smnt)

        self.nbas = len(self.Sijt[:, 0, 0])
        self.npot = len(self.Sijt[0, 0, :])
        self.b = np.zeros(self.npot)
        self.internal_b = np.ones_like(self.b)
        self.dm_aux = dm_aux
        self.dm_tar = dm_tar
        self.t_eig = self.t_ws = self.t_gd = self.t_hs = self.t_opt = 0.
        self.t_init = time.time()-t

    get_occ = scf.hf.get_occ
    make_rdm1 = scf.hf.SCF.make_rdm1

    run = run
    time_profile = time_profile
    info = info
    wyscf = wyscf

    def initialize(self, xbas=None):
        """Summary: Construct a fixed part of fock matrix F0

        .. math::

            F_{0} = T + V_{ext} + V_{h} + V_{g}

        """
        self.get_ovlp = lambda *args: self.S
        self.internal_b = np.ones_like(self.b)

        if self.dm_aux is None:
            self.dm_aux = self.dm_tar

        if xbas is None:
            xbas = self.tbas

        if self.guide is None:
            self.V0 = np.zeros_like(self.dm_tar)
        else:
            fac_faxc, dft_xc = util.parse_guide(self.guide)

            N = self.mol.nelectron

            self.xmol = df.make_auxmol(self.mol, auxbasis=xbas)
            dm_aux_ao = scf.addons.project_dm_nr2nr(self.xmol, self.dm_aux, self.mol)

            J_tar = scf.hf.get_jk(self.mol, dm_aux_ao)[0]
            VFA = -(1./N)*(J_tar)

            mydft = dft.RKS(self.mol)
            mydft.xc = dft_xc
            Vxcdft = mydft.get_veff(self.mol, dm=dm_aux_ao)

            self.V0 = fac_faxc * VFA + Vxcdft

            if self.Smnt is not None:
                #Generate potential matrix in target DM basis representation
                dm_aux_tbas = scf.addons.project_dm_nr2nr(self.xmol, self.dm_aux, self.tmol)
 
                self.V_tbas = self.tmol.intor('int1e_nuc')
                VFA_tbas = -(1./N)*scf.hf.get_jk(self.tmol, dm_aux_tbas)[0]
                mydft_tbas = dft.RKS(self.tmol)
                mydft_tbas.xc = dft_xc
                Vxcdft_tbas = mydft_tbas.get_veff(self.tmol, dm=dm_aux_tbas)
                self.V0_tbas = fac_faxc * VFA_tbas + Vxcdft_tbas

        self.F0 = self.T+self.V+self.V0

    def solve(self, b):
        """Summary: Under a given b vector, construct a fock matrix and diagonalize it

        F = F0 + V(b)

        FC = SCE

        resulting mo_coeff(C), mo_energy(E), and density matrix are stored as instance attributes
        """
        if not np.allclose(b, self.internal_b, rtol=1e-12, atol=1e-12):
            t = time.time()
            self.internal_b = b.copy()
            self.fock = self.F0+np.einsum('t,ijt->ij', b, self.Sijt) 
            self.mo_energy, self.mo_coeff = scf.hf.eig(self.fock, self.S)
            self.mo_occ = self.get_occ(self.mo_energy, self.mo_coeff)
            self.dm = self.make_rdm1(self.mo_coeff, self.mo_occ)
            self.t_eig += time.time()-t

            t = time.time()
            if kf_imported:
                if self.Smnt is None:
                    #AO basis for WY is same with AO basis of target density
                    self.grad = kspies_fort.einsum_ij_ijt_2t((self.dm-self.dm_tar), self.Sijt, self.nbas, self.npot)
                else:
                    self.grad = kspies_fort.einsum_ij_ijt_2t(self.dm, self.Sijt, self.nbas, self.npot)
                    self.grad -= kspies_fort.einsum_ij_ijt_2t(self.dm_tar, self.Smnt, len(self.Smnt[:, 0, 0]), self.npot)
            else:
                if self.Smnt is None:
                    self.grad = contract('ij,ijt->t', (self.dm-self.dm_tar), self.Sijt)
                else:
                    self.grad = contract('ij,ijt->t', self.dm, self.Sijt)
                    self.grad -= contract('ij,ijt->t', self.dm_tar, self.Smnt)

            self.t_gd += time.time()-t

    def eval_Ws(self, b):
        """Summary: Calculate objective function Ws under given b vector

        When mw.reg > 0, regularization is added
        """
        self.solve(b)
        t = time.time()

        Ws = np.einsum('ij,ji', self.T, self.dm)

        if self.Smnt is None:
            Ws += np.einsum('ij,ji', (self.V+self.V0), (self.dm-self.dm_tar))
        else:
            Ws += np.einsum('ij,ji', (self.V+self.V0), self.dm)
            Ws -= np.einsum('ij,ji', (self.V_tbas+self.V0_tbas), self.dm_tar)
        Ws += np.einsum('t,t', b, self.grad)
        self.Ws = Ws

        penelty = 0.
        if self.reg > 0:
            penelty = self.reg*self.Dvb(b)

        self.t_ws += time.time()-t
        return -(Ws-penelty)

    def eval_Gd(self, b):
        """Summary: -d(Ws)/b_t
        """
        self.solve(b)
        t = time.time()
        self.Gd = self.grad

        if self.reg > 0:
            self.Gd -= 4*self.reg*np.einsum('st,t->s', self.Tp, b)
        self.t_gd += time.time()-t
        return -self.Gd

    def eval_Hs(self, b):
        """Summary: -d^2(Ws)/(b_t)(b_u)
        """
        self.solve(b)
        t = time.time()

        nocc = self.mol.nelectron//2
        eia = self.mo_energy[:nocc, None] - self.mo_energy[None, nocc:]

        tol_zero = 1e+15
        with np.errstate(divide='ignore'):
            eia = np.nan_to_num(eia**-1, posinf=tol_zero, neginf=-tol_zero)**-1

        if kf_imported:
            self.Hs=kspies_fort.wy_hess(self.Sijt, self.mo_coeff, eia, nocc, self.nbas-nocc, self.npot)
        else:
            Siat = contract('mi,va,mvt->iat', self.mo_coeff[:,:nocc], self.mo_coeff[:,nocc:], self.Sijt)
            self.Hs= 4*contract('iau,iat,ia->ut', Siat, Siat, eia**-1)

        if self.reg > 0:
            self.Hs -= 4*self.reg*self.Tp
        self.t_hs += time.time()-t
        return -self.Hs

    def Dvb(self, b=None):
        """Summary: Calcuate the norm of the correction potential derivative

        .. math::

            \\| \\nabla v_C ( \\textbf{r} ) \\|^2 = \\int v_C ( \\textbf{r} ) \\nabla^2 v_C ( \\textbf{r} ) d \\textbf{r}

        """
        if b is None:
            b = self.b
        Dvb = 2*np.einsum('s,st,t', b, self.Tp, b, optimize=True)
        return Dvb

class UWY:
    """Summary: Perform WY calculation in unrestricted scheme

    .. _unrestrictedwy:

    Attributes:
        mol (object) : an instance of :class:`Mole`
        dm_tar (ndarray) : Density matrix of target density in atomic orbital basis representation
        pbas (dict or str) : Potential basis set for WY. If not given, same with atomic orbital basis
        Sijt (ndarray) : Three-center overlap integral. If not given, calculated analytically as overlap of atomic orbital basis
        dm_aux (ndarray) : Auxilary density matrix to construct density-dependent part of guiding potential. Default is dm_tar
        guide (str) : Guiding potential. Can be set as

            |  None : no guiding potential except external potential
            |  'faxc' : Exchange-correlation part of Fermi-Amaldi potential
            |  xc   : ks.xc attribute in pyscf DFT

        reg (float) : strength of regularization. Default is 0
        tol (float) : tolarance of optimization (maximum gradient element). Default is 1e-6
        method (str) : optimization algorithm used in SciPy. Can be (case insensitive) 'cg', 'bfgs', 'newton-cg',
                       'dogleg', 'trust-ncg', 'trust-krylov', 'trust-exact'. Defult is 'trust-exact'

    :ivar:
        * **converged** (bool) – optimization converged or not
        * **mo_energy** (ndarray) –  molecular orbital energies
        * **mo_occ** (ndarray) –  molecular orbital occupation numbers
        * **mo_coeff** (ndarray) – molecular orbital coefficients
        * **dm** (ndarray) – density matrix in atomic orbital basis representation
        * **b** (ndarray) – optimized linear combination coefficients.

    """
    def __init__(self, mol, dm_tar, pbas=None, Sijt=None, tbas=None, Smnt=None, dm_aux=None):
        t = time.time()
        basic(self, mol, pbas, Sijt, tbas, Smnt)

        self.nelec = [int((mol.nelectron+mol.spin)//2), int((mol.nelectron-mol.spin)//2)]
        self.nbas = len(self.Sijt[:, 0, 0])
        self.npot = len(self.Sijt[0, 0, :])
        self.b = np.zeros(2*self.npot)
        self.dm_aux = dm_aux
        self.dm_tar = dm_tar
        self.t_eig = self.t_ws = self.t_gd = self.t_hs = self.t_opt = 0.
        self.t_init = time.time()-t

    get_occ = scf.uhf.get_occ
    make_rdm1 = scf.uhf.UHF.make_rdm1
    spin_square = scf.uhf.UHF.spin_square

    run = run
    time_profile = time_profile
    info = info
    wyscf = wyscf

    def initialize(self, xbas=None):
        """Summary: Construct a fixed part of fock matrix F0

        .. math::

            F_{0} = T + V_{ext} + V_{h} + V_{g}

        """
        self.get_ovlp = lambda *args: self.S
        self.internal_b = np.ones_like(self.b)

        if self.dm_aux is None:
            self.dm_aux = self.dm_tar
        if xbas is None:
            xbas = self.tbas

        if self.guide is None:
            self.V0 = np.zeros_like(self.dm_tar)
        else:
            fac_faxc, dft_xc = util.parse_guide(self.guide)

            N = self.mol.nelectron
            self.xmol = df.make_auxmol(self.mol, auxbasis=xbas)
            dm_aux_ao = scf.addons.project_dm_nr2nr(self.xmol, self.dm_aux, self.mol)

            J_tar = scf.hf.get_jk(self.mol, dm_aux_ao)[0]
            VFA = -(1./N)*(J_tar[0]+J_tar[1])

            mydft = dft.UKS(self.mol)
            mydft.xc = dft_xc
            Vxcdft = mydft.get_veff(self.mol, dm=dm_aux_ao)

            if self.Smnt is not None:
                #Generate potential matrix in target DM basis representation
                dm_aux_tbas = scf.addons.project_dm_nr2nr(self.xmol, self.dm_aux, self.tmol)

                self.V_tbas = self.tmol.intor('int1e_nuc')
                J_tbas = scf.hf.get_jk(self.tmol, dm_aux_tbas)[0]
                VFA_tbas = -(1./N)*(J_tbas[0]+J_tbas[1])
                mydft_tbas = dft.UKS(self.tmol)
                mydft_tbas.xc = dft_xc
                Vxcdft_tbas = mydft_tbas.get_veff(self.tmol, dm=dm_aux_tbas)
                self.V0_tbas = fac_faxc * VFA_tbas + Vxcdft_tbas

            self.V0 = fac_faxc * np.array((VFA, VFA)) + Vxcdft

        self.F0 = (self.T+self.V+self.V0[0],
                   self.T+self.V+self.V0[1])


    def solve(self, b):
        """Summary: Under a given b vector, construct a fock matrix and diagonalize it

        F = F0 + V(b)

        FC = SCE

        resulting mo_coeff(C), mo_energy(E), and density matrix are stored as instance attributes
        """
        if not np.allclose(b, self.internal_b, rtol=1e-12, atol=1e-12):
            t = time.time()
            self.internal_b = b.copy()
            Fa = self.F0[0]+contract('t,ijt->ij', b[:self.npot], self.Sijt)
            Fb = self.F0[1]+contract('t,ijt->ij', b[self.npot:], self.Sijt)
            self.fock = ((Fa, Fb))
            e_a, c_a = scf.hf.eig(Fa, self.S)
            e_b, c_b = scf.hf.eig(Fb, self.S)

            self.mo_coeff = np.array((c_a, c_b))
            self.mo_energy = np.array((e_a, e_b))
            self.mo_occ = self.get_occ(self.mo_energy, self.mo_coeff)
            self.dm = self.make_rdm1(self.mo_coeff, self.mo_occ)
            self.t_eig += time.time()-t

            t = time.time()
            if kf_imported:
                if self.Smnt is None:
                    #AO basis for WY is same with AO basis of target density
                    self.grad_a = kspies_fort.einsum_ij_ijt_2t((self.dm[0]-self.dm_tar[0]), self.Sijt, self.nbas, self.npot)
                    self.grad_b = kspies_fort.einsum_ij_ijt_2t((self.dm[1]-self.dm_tar[1]), self.Sijt, self.nbas, self.npot)
                else:
                    self.grad_a = kspies_fort.einsum_ij_ijt_2t(self.dm[0], self.Sijt, self.nbas, self.npot)
                    self.grad_a -= kspies_fort.einsum_ij_ijt_2t(self.dm_tar[0], self.Smnt, len(self.Smnt[:, 0, 0]), self.npot)
                    self.grad_b = kspies_fort.einsum_ij_ijt_2t(self.dm[1], self.Sijt, self.nbas, self.npot)
                    self.grad_b -= kspies_fort.einsum_ij_ijt_2t(self.dm_tar[1], self.Smnt, len(self.Smnt[:, 0, 0]), self.npot)
            else:
                if self.Smnt is None:
                    self.grad_a = contract('ij,ijt->t', (self.dm[0]-self.dm_tar[0]), self.Sijt)
                    self.grad_b = contract('ij,ijt->t', (self.dm[1]-self.dm_tar[1]), self.Sijt)
                else:
                    self.grad_a = contract('ij,ijt->t', self.dm[0], self.Sijt)
                    self.grad_a -= contract('ij,ijt->t', self.dm_tar[0], self.Smnt)
                    self.grad_b = contract('ij,ijt->t', self.dm[1], self.Sijt)
                    self.grad_b -= contract('ij,ijt->t', self.dm_tar[1], self.Smnt)

            self.t_gd += time.time()-t

    def eval_Ws(self, b):
        """Summary: Calculate objective function Ws under given b vector

        When mw.reg > 0, regularization is added
        """
        self.solve(b)
        t = time.time()

        Ws = np.einsum('ij,ji', self.T, (self.dm[0]+self.dm[1]))

        if self.Smnt is None:
            Ws += np.einsum('ij,ji', self.V, (self.dm[0]+self.dm[1]-self.dm_tar[0]-self.dm_tar[1]))
            Ws += np.einsum('ij,ji', self.V0[0], (self.dm[0]-self.dm_tar[0]))
            Ws += np.einsum('ij,ji', self.V0[1], (self.dm[1]-self.dm_tar[1]))

        else:
            Ws += np.einsum('ij,ji', self.V, (self.dm[0]+self.dm[1]))
            Ws -= np.einsum('ij,ji', self.V_tbas, (self.dm_tar[0]+self.dm_tar[1]))
            Ws += np.einsum('ij,ji', self.V0[0], self.dm[0])
            Ws -= np.einsum('ij,ji', self.V0_tbas[0], self.dm_tar[0])
            Ws += np.einsum('ij,ji', self.V0[1], self.dm[1])
            Ws -= np.einsum('ij,ji', self.V0_tbas[1], self.dm_tar[1])

        Ws += np.einsum('t,t', b[:self.npot], self.grad_a)
        Ws += np.einsum('t,t', b[self.npot:], self.grad_b)
        self.Ws = Ws

        penelty = 0.
        if self.reg > 0: #Regularization on
            penelty = self.reg*self.Dvb(b)

        self.t_ws += time.time()-t
        return -(Ws-penelty)

    def eval_Gd(self, b):
        """Summary: -d(Ws)/b_t
        """
        self.solve(b)
        t = time.time()
        self.Ga = self.grad_a
        self.Gb = self.grad_b

        if self.reg > 0: #Regularization on
            self.Ga -= 2*self.reg*np.einsum('st,t->s', self.Tp, b[:self.npot])
            self.Gb -= 2*self.reg*np.einsum('st,t->s', self.Tp, b[self.npot:])

        self.Gd = np.concatenate((self.Ga, self.Gb), axis=None)
        self.t_gd += time.time()-t
        return -self.Gd

    def eval_Hs(self, b):
        """Summary: -d^2(Ws)/(b_t)(b_u)
        """
        self.solve(b)
        t = time.time()

        n_a, n_b = self.nelec[0], self.nelec[1]
        mo_a, mo_b = self.mo_coeff
        eia_a = self.mo_energy[0][:n_a, None] - self.mo_energy[0][None, n_a:]
        eia_b = self.mo_energy[1][:n_b, None] - self.mo_energy[1][None, n_b:]

        tol_zero = 1e+15
        with np.errstate(divide='ignore'):
            eia_a = np.nan_to_num(eia_a**-1, posinf=tol_zero, neginf=-tol_zero)**-1
            eia_b = np.nan_to_num(eia_b**-1, posinf=tol_zero, neginf=-tol_zero)**-1

        if kf_imported:
            self.Ha = .5*kspies_fort.wy_hess(self.Sijt, mo_a, eia_a, n_a, self.nbas-n_a, self.npot)
            self.Hb = .5*kspies_fort.wy_hess(self.Sijt, mo_b, eia_b, n_b, self.nbas-n_b, self.npot)
        else:
            Siat_a = contract('mi,va,mvt->iat', mo_a[:,:n_a], mo_a[:,n_a:], self.Sijt)
            Siat_b = contract('mi,va,mvt->iat', mo_b[:,:n_b], mo_b[:,n_b:], self.Sijt)
            self.Ha = 2*contract('iau,iat,ia->ut', Siat_a, Siat_a, eia_a**-1)
            self.Hb = 2*contract('iau,iat,ia->ut', Siat_b, Siat_b, eia_b**-1)
        self.Hs = np.block([[self.Ha, np.zeros((self.npot, self.npot))],
                            [np.zeros((self.npot, self.npot)), self.Hb]])

        if self.reg > 0:
            self.Hs[self.npot:,self.npot:] -= 2*self.reg*self.Tp
            self.Hs[:self.npot,:self.npot] -= 2*self.reg*self.Tp
        self.t_hs += time.time()-t
        return -self.Hs

    def Dvb(self, b=None):
        """Summary: Calcuate the norm of correction potential derivative

        .. math::

            \\| \\nabla v_C ( \\textbf{r} ) \\| = \\frac{1}{2} \\sum_{\\sigma} \\int v_{C,\\sigma} ( \\textbf{r} ) \\nabla^2 v_{C,\\sigma} ( \\textbf{r} ) d \\textbf{r}

        """
        if b is None:
            ba = self.b[:len(self.b)//2]
            bb = self.b[len(self.b)//2:]
        else:
            ba = b[:len(b)//2]
            bb = b[len(b)//2:]
        Dvb = contract('s,st,t', ba, self.Tp, ba)
        Dvb += contract('s,st,t', bb, self.Tp, bb)
        return Dvb



import setuptools
from numpy.distutils.core import setup, Extension

print(setuptools.find_packages())

lib = Extension(
        name='kspies.kspies_fort',
        sources=['kspies/kspies_fort.f90'],
        library_dirs=['/usr/lib/'],
        include_dirs=['/usr/include'],
        libraries=['lapack']
        )

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="kspies", # Replace with your own username
    version="1.0.2",
    author="Seungsoo Nam,  Ryan J. McCarty, Hansol Park, Eunji Sim",
    author_email="skaclitz@yonsei.ac.kr",
    description="This is a python based Kohn-Sham Inversion Evaluation Software package for use with pySCF.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://ssnam92.github.io/KSPies/",
    packages=setuptools.find_packages(),
    ext_modules = [lib],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
        "Development Status :: 4 - Beta",
    ],
    python_requires='>=3.6',
)
